//	CZT[ BMP180CuiW[ GY-68j
//	wt]xH쎺	@jsdiy	http://jsdiy.webcrow.jp

/*
yXVz
2016/10/29	v1.00	

2020/04/13	v2.00
	EReadRegValue()ATWI_ReadByte()twilib-v2.00̎dlɍ킹
	E\̖ύX
*/

#include <avr/io.h>
#include <util/delay.h>
#include "twilib.h"
#include "bmp180lib.h"

//debug	(0 or 1)
#define _BOSCH_DATASHEET_SAMPLE_	0

//BMP180̃X[uAhX(7bit)
#define	BMP180_SLAVE_ADDRESS	0x77

//Calibration coefficients
typedef	struct
{
	int16_t		ac1;
	int16_t		ac2;
	int16_t		ac3;
	uint16_t	ac4;
	uint16_t	ac5;
	uint16_t	ac6;
	int16_t		b1;
	int16_t		b2;
	int16_t		mb;
	int16_t		mc;
	int16_t		md;
	//
	int32_t		b5;
	uint8_t		oss;
}
BMP180_TCalParam;

//Calibration coefficients
static	BMP180_TCalParam	calParam;

//֐
static	void	WriteRegValue(uint8_t regAddr, uint8_t dataByte);
static	void	ReadRegValue(uint8_t startAddr, uint8_t readLength, uint8_t* retBuf);
static	void	DelayMSec(uint8_t msec);
static	void	SoftReset(void);
static	uint8_t	ComCheck(void);
static	void	ReadCalData(void);
static	int32_t	ReadUncomTemperature(void);
static	int32_t	ReadUncomPressure(BMP180_EOverSamplingSetting oss);
static	void	CalcTemperature(int32_t ut, BMP180_TDecimal* retValue);
static	void	CalcPressure(int32_t up, BMP180_TDecimal* retValue);


//WX^AhXXXhɃf[^0xYY
static	void	WriteRegValue(uint8_t regAddr, uint8_t dataByte)
{
	TWI_Start();
	TWI_SlaveAddrW(BMP180_SLAVE_ADDRESS);
	TWI_WriteByte(regAddr);
	TWI_WriteByte(dataByte);
	TWI_Stop();
}

//wAhXɊi[ꂽlǂ
static	void	ReadRegValue(uint8_t startAddr, uint8_t readLength, uint8_t* retBuf)
{
	TWI_Start();
	TWI_SlaveAddrW(BMP180_SLAVE_ADDRESS);
	TWI_WriteByte(startAddr);
	TWI_Stop();
	
	TWI_Start();
	TWI_SlaveAddrR(BMP180_SLAVE_ADDRESS);
	for (uint8_t i = 0; i < readLength; i++)
	{
		//TWI_ReadByte(i < (readLength - 1) ? 1 : 0, retBuf + i);	//twilib-v1.20
		TWI_ReadByte(i == (readLength - 1) ? 1 : 0, retBuf + i);	//twilib-v2.00cTWI_ReadByte()̈ύXĂI
	}
	TWI_Stop();
}

//1msPʂ̎ԑ҂
static	void	DelayMSec(uint8_t msec)
{
	for (uint8_t i = 0; i < msec; i++) { _delay_ms(1); }
}

//
uint8_t	BMP180_Init(void)
{
	SoftReset();
	ReadCalData();
	return ComCheck();	//̐ۂ̕ԓ˂
}

//\tgZbg
static	void	SoftReset(void)
{
	//WX^AhXE0h0xB6
	WriteRegValue(0xE0, 0xB6);

	//uStart-up time after power-up, before first communication : Min.10msv
	//p[IZbg̑҂ԂQlɂB̂]TB
	_delay_ms(100);
}

//ʐMmF
static	uint8_t	ComCheck(void)
{
	//WX^AhXD0hǂ0x55BMP180ƒʐMĂ邱ƂɂȂ
	uint8_t	data;
	ReadRegValue(0xD0, 1, &data);
	
	return (data == 0x55);
}

//read calibration data
static	void	ReadCalData(void)
{
	const uint8_t CALPARAM_LENGTH = 0xBF - 0xAA + 1;	//AhXAAhBFh܂ł̃oCg
	uint8_t	buf[CALPARAM_LENGTH];
	
	//calibration dataǂ
	ReadRegValue(0xAA, CALPARAM_LENGTH, buf);

	//obt@̒lHč\̂֊i[
	uint8_t i = 0;
	calParam.ac1 = (int16_t)(((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1]);	i += 2;
	calParam.ac2 = (int16_t)(((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1]);	i += 2;
	calParam.ac3 = (int16_t)(((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1]);	i += 2;
	calParam.ac4 =			((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1];		i += 2;
	calParam.ac5 =			((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1];		i += 2;
	calParam.ac6 =			((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1];		i += 2;
	calParam.b1 = (int16_t)(((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1]);	i += 2;
	calParam.b2 = (int16_t)(((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1]);	i += 2;
	calParam.mb = (int16_t)(((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1]);	i += 2;
	calParam.mc = (int16_t)(((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1]);	i += 2;
	calParam.md = (int16_t)(((uint16_t)buf[i] << 8) | (uint16_t)buf[i + 1]);	i += 2;

#if (_BOSCH_DATASHEET_SAMPLE_)
	//BOSCH datasheet sample
	calParam.ac1 = 408;
	calParam.ac2 = -72;
	calParam.ac3 = -14383;
	calParam.ac4 = 32741;
	calParam.ac5 = 32757;
	calParam.ac6 = 23153;
	calParam.b1 = 6190;
	calParam.b2 = 4;
	calParam.mb = -32768;
	calParam.mc = -8711;
	calParam.md = 2868;
#endif
}

//read uncompensated temperature value
static	int32_t	ReadUncomTemperature(void)
{
	uint8_t buf[2];
	
	//WX^AhXF4h0x2EA4.5ms҂ĂAWX^AhXF6h(MSB),F7h(LSB)ǂ
	WriteRegValue(0xF4, 0x2E);
	DelayMSec(5);
	ReadRegValue(0xF6, 2, buf);
	int32_t ut = (int32_t)(((uint32_t)buf[0] << 8) | (uint32_t)buf[1]);

#if (_BOSCH_DATASHEET_SAMPLE_)
	//BOSCH datasheet sample
	ut = 27898;
#endif
	
	return ut;
}

//read uncompensated pressure value
static	int32_t	ReadUncomPressure(BMP180_EOverSamplingSetting oss)
{
	uint8_t ctrlRegVal, waitMSec;
	uint8_t buf[3];
	
#if (_BOSCH_DATASHEET_SAMPLE_)
	//BOSCH datasheet sample
	oss = BMP180_OSS_UltraLowPower;
#endif

	//Control register value
	switch (oss)
	{
	case BMP180_OSS_UltraLowPower:			ctrlRegVal = 0x34;	waitMSec = 5;	break;	//4.5ms
	case BMP180_OSS_Standard:				ctrlRegVal = 0x74;	waitMSec = 8;	break;	//7.5ms
	case BMP180_OSS_HighResolution:			ctrlRegVal = 0xB4;	waitMSec = 14;	break;	//13.5ms
	case BMP180_OSS_UltraHighResolution:	ctrlRegVal = 0xF4;	waitMSec = 26;	break;	//25.5ms
	default:	ctrlRegVal = 0x00;	waitMSec = 0;	break;
	}
	
	//WX^AhXF4hɃRg[lAK̎Ԃ҂AWX^AhXF6h(MSB),F7h(LSB),F8h(XLSB)ǂ
	WriteRegValue(0xF4, ctrlRegVal);
	DelayMSec(waitMSec);
	ReadRegValue(0xF6, 3, buf);

	//UP = (msb<<16 + lsb<<8 + xlsb)>>(8-oss)
	int32_t up = (int32_t)((((uint32_t)buf[0] << 16) | ((uint32_t)buf[1] << 8) | (uint32_t)buf[2]) >> (8 - oss));

#if (_BOSCH_DATASHEET_SAMPLE_)
	//BOSCH datasheet sample
	up = 23843;
#endif

	calParam.oss = (uint8_t)oss;
	
	return up;
}

//calculate true temperature
static	void	CalcTemperature(int32_t ut, BMP180_TDecimal* retValue)
{
	int32_t	x1, x2, t;	//t0.1xPʁiFt=153  15.3xj
	
	x1 = ((ut - (int32_t)calParam.ac6) * (int32_t)calParam.ac5) >> 15;
	x2 = ((int32_t)calParam.mc << 11) / (x1 + (int32_t)calParam.md);
	calParam.b5 = x1 + x2;
	t = (calParam.b5 + 8) >> 4;
	
	retValue->integer = (int16_t)(t / 10);
	if (t < 0) { t = -t; }
	retValue->decimal = (uint8_t)(t % 10);
}

//calculate true pressure
static	void	CalcPressure(int32_t up, BMP180_TDecimal* retValue)
{
	int32_t	x1, x2, x3, b6, b3, p;
	uint32_t	b4, b7;

	b6 = calParam.b5 - 4000;
	x1 = (calParam.b2 * ((b6 * b6) >> 12)) >> 11;
	x2 = (calParam.ac2 * b6) >> 11;
	x3 = x1 + x2;
	b3 = ((((int32_t)calParam.ac1 * 4 + x3) << calParam.oss) + 2) / 4;
	x1 = (calParam.ac3 * b6) >> 13;
	x2 = (calParam.b1 * ((b6 * b6) >> 12)) >> 16;
	x3 = ((x1 + x2) + 2) >> 2;
	b4 = ((uint32_t)calParam.ac4 * (uint32_t)(x3 + 32768)) >> 15;
	b7 = (uint32_t)(up - b3) * (50000 >> calParam.oss);
	if (b7 < 0x80000000)
	{
		p = (int32_t)((b7 * 2) / b4);
	}
	else
	{
		p = (int32_t)((b7 / b4) * 2);
	}
	x1 = (p >> 8) * (p >> 8);
	x1 = (x1 * 3038) >> 16;
	x2 = (-7357 * p) >> 16;
	p = p + ((x1 + x2 + 3791) >> 4);
	
	//Pa -> hPa		p65535𒴂邽
	retValue->integer = (int16_t)(p / 100);
	retValue->decimal = (uint8_t)(p % 100);	//2ʂ܂
	
	//1ʂ܂Łi2ʂľܓj
	if (95 <= retValue->decimal){ retValue->integer++;	retValue->decimal = 0; }
	else { retValue->decimal = (retValue->decimal + 5) / 10; }
}

//x̑l擾
void	BMP180_GetTemperature(BMP180_TDecimal* retValue)
{
	int32_t	ut = ReadUncomTemperature();
	CalcTemperature(ut, retValue);
}

//C̑l擾
void	BMP180_GetPressure(BMP180_TDecimal* retValue, BMP180_EOverSamplingSetting oss)
{
	int32_t	up = ReadUncomPressure(oss);
	CalcPressure(up, retValue);
}
